// ======================================================================
//                     ORBITER SOFTWARE DEVELOPMENT KIT
//                  Copyright (C) 2001-2004 Martin Schweiger
//                           All rights reserved
// OrbiterAPI.h
// ORBITER Application Programming Interface (OAPI)
// ======================================================================

#ifndef __ORBITERAPI_H
#define __ORBITERAPI_H

#include <windows.h>
#include <fstream.h>
#include <float.h>
#include <math.h>

// Assumes MS VC++ compiler. Modify these statements for other compilers
#define DLLEXPORT __declspec(dllexport)
#define DLLIMPORT __declspec(dllimport)
#define DLLCLBK extern "C" __declspec(dllexport)

#ifdef OAPI_IMPLEMENTATION
#define OAPIFUNC DLLEXPORT
#else
#define OAPIFUNC DLLIMPORT
#endif

// ======================================================================
// Some useful constants
// ======================================================================

const double PI   = 3.14159265358979;
const double PI05 = 1.57079632679490;
const double RAD  = PI/180.0;
const double DEG  = 180.0/PI;
const double G    = 9.81;         // gravitational acceleration [m/s^2] at Earth mean radius
const double ATMP = 101.4e3;      // atmospheric pressure [Pa] at Earth sea level
const double ATMD = 1.293;        // atmospheric density [kg/m^3] at Earth sea level

// ======================================================================
// API data types
// ======================================================================

class VESSEL;

typedef void *OBJHANDLE;
// Handle for objects (vessels, stations, planets)

typedef void *VISHANDLE;
// Handle for visuals

typedef void *MESHHANDLE;
// Handle for meshes

typedef void *SURFHANDLE;
// Handle for bitmap surfaces (panels and panel items)

typedef void *FILEHANDLE;

typedef void *THRUSTER_HANDLE;
typedef void *THGROUP_HANDLE;
// Handle for (logical) thrusters and thruster groups

typedef void *PROPELLANT_HANDLE;
// Handle for propellant resource

typedef void *PSTREAM_HANDLE;
// Handle for particle streams

typedef void *DOCKHANDLE;
// Handle for vessel docking ports

typedef void *ATTACHMENTHANDLE;
// Handle vor vessel passive attachment points

typedef void *AIRFOILHANDLE;
// Handle for vessel airfoils

typedef void *NAVHANDLE;
// Handle for a navigation radio transmitter (VOR, ILS, IDS, XPDR)

typedef void *ANIMATIONCOMPONENT_HANDLE;

typedef union {      // 3 vector
	double data[3];
	struct { double x, y, z; };
} VECTOR3;

typedef union {      // 3x3 matrix
	double data[9];
	struct { double m11, m12, m13, m21, m22, m23, m31, m32, m33; };
} MATRIX3;

typedef union {      // 4x4 matrix
	double data[16];
	struct { double m11, m12, m13, m14, m21, m22, m23, m24, m31, m32, m33, m34, m41, m42, m43, m44; };
} MATRIX4;

typedef struct {    // vertex definition including normals and texture coordinates
	float x, y, z;     // position
	float nx, ny, nz;  // normal
	float tu, tv;      // texture coordinates
} NTVERTEX;

typedef struct {    // mesh group definition
	NTVERTEX *Vtx;     // vertex list
	WORD *Idx;         // index list
	DWORD nVtx;        // vertex count
	DWORD nIdx;        // index count
	DWORD MtrlIdx;     // material index (>= 1, 0=none)
	DWORD TexIdx;      // texture index (>= 1, 0=none)
	DWORD UsrFlag;     // user-defined flag
	WORD zBias;        // z bias
	WORD Flags;        // internal flags
} MESHGROUP;

typedef struct {       // Orbital elements
	double a;          //     semi-major axis [m]
	double e;          //     eccentricity
	double i;          //     inclination [rad]
	double theta;      //     longitude of ascending node [rad]
	double omegab;     //     longitude of periapsis [rad]
	double L;          //     mean longitude at epoch
} ELEMENTS;

typedef struct {       // Planetary atmospheric constants
	double p0;         //     pressure at mean radius ('sea level') [Pa]
	double rho0;       //     density at mean radius
	double R;          //     specific gas constant [J/(K kg)]
	double gamma;      //     ratio of specific heats, c_p/c_v
	double C;          //     exponent for pressure equation (temporary)
	double O2pp;       //     partial pressure of oxygen
	double altlimit;   //     atmosphere altitude limit [m]
	double radlimit;   //     radius limit (altlimit + mean radius)
	double horizonalt; //     horizon rendering altitude
	VECTOR3 color0;    //     sky colour at sea level during daytime
} ATMCONST;

typedef struct {       // Atmospheric parameters
	double T;          //     temperature [K]
	double p;          //     pressure [Pa]
	double rho;        //     density [kg/m^3]
} ATMPARAM;

typedef struct {       // Engine status
	double main;       //     -1 (full retro) .. +1 (full main)
	double hover;      //     0 .. +1 (full hover)
	int attmode;       //     0=rotation, 1=translation
} ENGINESTATUS;

typedef struct {       // Particle stream parameters
	DWORD flags;       //     streamspec bitflags
	double srcsize;    //     particle size at creation [m]
	double srcrate;    //     average particle creation rate [Hz]
	double v0;         //     emission velocity [m/s]
	double srcspread;  //     velocity spread during creation
	double lifetime;   //     average particle lifetime
	double growthrate; //     particle growth rate [m/s]
	double atmslowdown;//     slowdown rate in atmosphere
	enum LTYPE { EMISSIVE, DIFFUSE } ltype; // render lighting method
	enum LEVELMAP { LVL_FLAT, LVL_LIN, LVL_SQRT, LVL_PLIN, LVL_PSQRT } levelmap; // mapping from level to alpha
	double lmin, lmax; //     min and max levels for level PLIN and PSQRT mapping types
	enum ATMSMAP { ATM_FLAT, ATM_PLIN, ATM_PLOG } atmsmap;    // mapping from atmospheric params to alpha
	double amin, amax; //     min and max densities for atms PLIN mapping
	SURFHANDLE tex;    //     particle texture handle (NULL for default)
} PARTICLESTREAMSPEC;

typedef struct {
	VECTOR3 rpos;      //   position relative to rbody (ecliptic frame)
	VECTOR3 rvel;      //   velocity relative to rbody (ecliptic frame)
	VECTOR3 vrot;      //   rotation velocity about principal axes
	VECTOR3 arot;      //   vessel orientation against ecliptic frame
	double fuel;       //   fuel level [0..1]
	double eng_main;   //   main engine setting [-1..1]
	double eng_hovr;   //   hover engine setting [0..1]
	OBJHANDLE rbody;   //   handle of reference body
	OBJHANDLE base;    //   handle of docking or landing target
	int port;          //   designated docking or landing port
	int status;        //   0=freeflight, 1=landed, 2=taxiing, 3=docked, 99=undefined

	// reserved data structures for future use
	VECTOR3 vdata[10];  //   vdata[0]: contains x=long,y=lat of landing site and z=vessel orientation if status==1
	double  fdata[10];
	DWORD   flag[10];
} VESSELSTATUS;

typedef struct {      // vessel status: interface version 2
	DWORD version;      // version id
	DWORD flag;         // bitflags
	OBJHANDLE rbody;    // handle of reference body
	OBJHANDLE base;     // handle of docking or landing target
	int port;           // designated docking or landing port
	int status;         // 0=freeflight, 1=landed, 3=docked to station (obsolete)
	VECTOR3 rpos;       // position relative to rbody (ecliptic frame)
	VECTOR3 rvel;       // velocity relative to rbody (ecliptic frame)
	VECTOR3 vrot;       // rotation velocity about principal axes
	VECTOR3 arot;       // vessel orientation against ecliptic frame
	double surf_lng;    // longitude of landing position  |
	double surf_lat;    // latitude of landing position   | only used when landed
	double surf_hdg;    // heading of landed vessel       |
	DWORD nfuel;        // length of propellant level list
	struct FUELSPEC {
		DWORD idx;      // propellant index
		double level;   // propellant level
	} *fuel;            // propellant list
	DWORD nthruster;    // length of thruster level list
	struct THRUSTSPEC {
		DWORD idx;      // thruster index 
		double level;   // thruster level
	} *thruster;        // thruster list
	DWORD ndockinfo;    // length of dock info list
	struct DOCKINFOSPEC {
		DWORD idx;      // docking port index
		DWORD ridx;     // docking port index of docked vessel
		OBJHANDLE rvessel; // docked vessel
	} *dockinfo;        // dock info list
	DWORD xpdr;         // transponder setting (in 0.05kHz steps from 108.00)
} VESSELSTATUS2;

// VESSELSTATUSx bitflags
#define VS_FUELRESET    0x00000001
#define VS_FUELLIST     0x00000002
#define VS_THRUSTRESET  0x00000004
#define VS_THRUSTLIST   0x00000008
#define VS_DOCKINFOLIST 0x00000010

#pragma pack(push,1)
typedef struct {
	union {
		struct {
			VECTOR3 ref;
			VECTOR3 axis;
			float angle;
		} rotparam;
		struct {
			VECTOR3 shift;
		} transparam;
		struct {
			VECTOR3 scale;
		} scaleparam;
	} P;
	int nmesh;
	int ngrp;
	enum { TRANSLATE, ROTATE, SCALE } transform;
} MESHGROUP_TRANSFORM;
#pragma pack(pop)

typedef struct {
	UINT *grp;
	UINT ngrp;
	double state0;
	double state1;
	MESHGROUP_TRANSFORM trans;
} ANIMCOMP;

// Transformation types for animations

// Transformation base class
class MGROUP_TRANSFORM {
public:
	MGROUP_TRANSFORM () : mesh(0), grp(0), ngrp(0) {}
	MGROUP_TRANSFORM (UINT _mesh, UINT *_grp, UINT _ngrp)
		: mesh(_mesh), grp(_grp), ngrp(_ngrp) {}
	enum TYPE { NULLTRANSFORM, ROTATE, TRANSLATE, SCALE };
	virtual TYPE Type() const { return NULLTRANSFORM; }

	UINT mesh;
	UINT *grp;
	UINT ngrp;
};

// Rotation
class MGROUP_ROTATE: public MGROUP_TRANSFORM {
public:
	MGROUP_ROTATE (UINT _mesh, UINT *_grp, UINT _ngrp, const VECTOR3 &_ref, const VECTOR3 &_axis, float _angle)
		: MGROUP_TRANSFORM (_mesh, _grp, _ngrp), ref(_ref), axis(_axis), angle(_angle) {}
	TYPE Type() const { return ROTATE; }

	VECTOR3 ref;
	VECTOR3 axis;
	float angle;
};

// Translation
class MGROUP_TRANSLATE: public MGROUP_TRANSFORM {
public:
	MGROUP_TRANSLATE (UINT _mesh, UINT *_grp, UINT _ngrp, const VECTOR3 &_shift)
		: MGROUP_TRANSFORM (_mesh, _grp, _ngrp), shift(_shift) {}
	TYPE Type() const { return TRANSLATE; }

	VECTOR3 shift;
};

// Scaling
class MGROUP_SCALE: public MGROUP_TRANSFORM {
public:
	MGROUP_SCALE (UINT _mesh, UINT *_grp, UINT _ngrp, const VECTOR3 &_ref, const VECTOR3 &_scale)
		: MGROUP_TRANSFORM (_mesh, _grp, _ngrp), ref(_ref), scale(_scale) {}
	TYPE Type() const { return SCALE; }

	VECTOR3 ref;
	VECTOR3 scale;
};

#define LOCALVERTEXLIST ((UINT)(-1))
#define MAKEGROUPARRAY(x) ((UINT*)x)

typedef struct {
	RECT pos;
	int nbt_left, nbt_right;
	int bt_yofs, bt_ydist;
} MFDSPEC;

typedef struct {
	DWORD nmesh, ngroup;
} VCMFDSPEC;

typedef struct {
	DWORD nmesh, ngroup;
	VECTOR3 hudcnt;
	double size;
} VCHUDSPEC;

typedef struct {
	int W, H;
	int CX, CY;
	double Scale;
	int Markersize;
} HUDPAINTSPEC;

typedef struct {
	char *name;
	DWORD key;
	int (*msgproc)(UINT,UINT,WPARAM,LPARAM);
} MFDMODESPEC;

#pragma pack(push,1)
typedef struct {
	const char *line1, *line2;
	char selchar;
} MFDBUTTONMENU;
#pragma pack(pop)

typedef enum {
	ENGINE_MAIN,
	ENGINE_RETRO,
	ENGINE_HOVER,
	ENGINE_ATTITUDE
} ENGINETYPE;

typedef enum {
	EXHAUST_MAIN,
	EXHAUST_RETRO,
	EXHAUST_HOVER,
	EXHAUST_CUSTOM
} EXHAUSTTYPE;

typedef enum {
	THGROUP_MAIN,
	THGROUP_RETRO,
	THGROUP_HOVER,
	THGROUP_ATT_PITCHUP,
	THGROUP_ATT_PITCHDOWN,
	THGROUP_ATT_YAWLEFT,
	THGROUP_ATT_YAWRIGHT,
	THGROUP_ATT_BANKLEFT,
	THGROUP_ATT_BANKRIGHT,
	THGROUP_ATT_RIGHT,
	THGROUP_ATT_LEFT,
	THGROUP_ATT_UP,
	THGROUP_ATT_DOWN,
	THGROUP_ATT_FORWARD,
	THGROUP_ATT_BACK,
	THGROUP_USER = 0x40
} THGROUP_TYPE;

typedef enum {
	ATTMODE_DISABLED,
	ATTMODE_ROT,
	ATTMODE_LIN,
} ATTITUDEMODE;

typedef double (*LiftCoeffFunc)(double aoa); // obsolete

typedef void (*AirfoilCoeffFunc)(
	double aoa, double M, double Re,
	double *cl, double *cm, double *cd);
// template for aerodynamic coefficients callback function
// Used to calculate lift coefficient cl, moment coefficient
// cm, and drag coefficient cd as a function of angle of attack
// aoa, Mach number M, and Reynolds number Re

typedef enum {
	LIFT_VERTICAL,
	LIFT_HORIZONTAL
} AIRFOIL_ORIENTATION;

typedef enum {
	AIRCTRL_ELEVATOR,
	AIRCTRL_RUDDER,
	AIRCTRL_AILERON,
	AIRCTRL_FLAP,
	AIRCTRL_ELEVATORTRIM,
	AIRCTRL_RUDDERTRIM
} AIRCTRL_TYPE;

#define AIRCTRL_AXIS_AUTO        0
#define AIRCTRL_AXIS_YPOS        1
#define AIRCTRL_AXIS_YNEG        2
#define AIRCTRL_AXIS_XPOS        3
#define AIRCTRL_AXIS_XNEG        4

#define NAVMODE_KILLROT          1
#define NAVMODE_HLEVEL           2
#define NAVMODE_PROGRADE         3
#define NAVMODE_RETROGRADE       4
#define NAVMODE_NORMAL           5
#define NAVMODE_ANTINORMAL       6
#define NAVMODE_HOLDALT          7

#define MANCTRL_ATTMODE          0
#define MANCTRL_REVMODE          1
#define MANCTRL_ROTMODE          2
#define MANCTRL_LINMODE          3
#define MANCTRL_ANYMODE          4

#define MANCTRL_KEYBOARD         0
#define MANCTRL_JOYSTICK         1
#define MANCTRL_ANYDEVICE        2

#define COCKPIT_GENERIC          1
#define COCKPIT_PANELS           2
#define COCKPIT_VIRTUAL          3

#define CAM_COCKPIT              0
#define CAM_TARGETRELATIVE       1
#define CAM_ABSDIRECTION         2
#define CAM_GLOBALFRAME          3
#define CAM_TARGETTOOBJECT       4
#define CAM_TARGETFROMOBJECT     5
#define CAM_GROUNDOBSERVER       6

#define RCS_NONE                 0
#define RCS_ROT                  1
#define RCS_LIN                  2

#define HUD_NONE                 0
#define HUD_ORBIT                1
#define HUD_SURFACE              2
#define HUD_DOCKING              3

#define MFD_NONE                 0
#define MFD_ORBIT                1
#define MFD_SURFACE              2
#define MFD_MAP                  3
#define MFD_HSI                  4
#define MFD_LANDING              5
#define MFD_DOCKING              6
#define MFD_OPLANEALIGN          7
#define MFD_OSYNC                8
#define MFD_TRANSFER             9
#define MFD_COMMS               10
#define MFD_USERTYPE            64
#define BUILTIN_MFD_MODES       10

#define MAXMFD                   5
#define MFD_LEFT                 0
#define MFD_RIGHT                1
#define MFD_USER1                2
#define MFD_USER2                3
#define MFD_USER3                4

#define RADIOSOURCE_NONE         0
#define RADIOSOURCE_VOR          1
#define RADIOSOURCE_VTOL_ILS     2

#define PANEL_LEFT               0
#define PANEL_RIGHT              1
#define PANEL_UP                 2
#define PANEL_DOWN               3

#define PANEL_REDRAW_NEVER       0
#define PANEL_REDRAW_ALWAYS      1
#define PANEL_REDRAW_MOUSE       2
#define PANEL_REDRAW_INIT        3
#define PANEL_REDRAW_USER        4

#define PANEL_MOUSE_IGNORE       0
#define PANEL_MOUSE_LBDOWN       1
#define PANEL_MOUSE_RBDOWN       2
#define PANEL_MOUSE_LBUP         4
#define PANEL_MOUSE_RBUP         8
#define PANEL_MOUSE_LBPRESSED   16
#define PANEL_MOUSE_RBPRESSED   32
#define PANEL_MOUSE_DOWN         3
#define PANEL_MOUSE_UP          12
#define PANEL_MOUSE_PRESSED     48

#define PANEL_MAP_NONE           0
#define PANEL_MAP_BACKGROUND     1
#define PANEL_MAP_CURRENT        2
#define PANEL_MAP_BGONREQUEST    3

#define PANEL_ATTACH_BOTTOM    0x0001
#define PANEL_ATTACH_TOP       0x0002
#define PANEL_ATTACH_LEFT      0x0004
#define PANEL_ATTACH_RIGHT     0x0008
#define PANEL_MOVEOUT_BOTTOM   0x0010
#define PANEL_MOVEOUT_TOP      0x0020
#define PANEL_MOVEOUT_LEFT     0x0040
#define PANEL_MOVEOUT_RIGHT    0x0080

#define SURF_NO_CK         0xFFFFFFFF
#define SURF_PREDEF_CK     0xFFFFFFFE

#define SURF_NO_ROTATION ((DWORD)-1)
#define SURF_HMIRROR     ((DWORD)-2)
#define SURF_VMIRROR     ((DWORD)-3)
#define SURF_ROTATE_90   ((DWORD)-4)
#define SURF_ROTATE_180  ((DWORD)-5)
#define SURF_ROTATE_270  ((DWORD)-6)

// Custom MFD message identifiers
#define OAPI_MSG_MFD_OPENED    1
#define OAPI_MSG_MFD_CLOSED    2
#define OAPI_MSG_MFD_UPDATE    3

#define MESHVIS_NEVER          0
#define MESHVIS_EXTERNAL       1
#define MESHVIS_COCKPIT        2
#define MESHVIS_ALWAYS         3
#define MESHVIS_VC             4

const UINT ALLDOCKS = (UINT)-1;

typedef int (*KeyFunc)(const char *keybuf);

// ======================================================================
// class MFD
// ======================================================================

class DLLEXPORT MFD {
public:
	MFD (DWORD w, DWORD h, VESSEL *vessel);
	virtual ~MFD();
	virtual void Update (HDC hDC) = 0;
	void InvalidateDisplay ();
	void Title (HDC hDC, const char *title) const;
	HPEN SelectDefaultPen (HDC hDC, DWORD i) const;
	HFONT SelectDefaultFont (HDC hDC, DWORD i) const;
	virtual bool ConsumeKeyBuffered (DWORD key) { return false; }
	virtual bool ConsumeKeyImmediate (char *kstate) { return false; }
	virtual bool ConsumeButton (int bt, int event) { return false; }
	virtual char *ButtonLabel (int bt) { return 0; }
	virtual int  ButtonMenu (const MFDBUTTONMENU **menu) const { return 0; }
	virtual void WriteStatus (FILEHANDLE scn) const {}
	virtual void ReadStatus (FILEHANDLE scn) {}
	virtual void StoreStatus (void) const {}
	virtual void RecallStatus (void) {}

protected:
	DWORD W, H;   // width and height of MFD display area (pixel)
	DWORD cw, ch; // character width, height of standard MFD font 0 (pixel)
	VESSEL *pV;   // pointer to vessel interface

private:
	friend class Instrument_User;  // Orbiter private class
	Instrument_User *instr;
};

// ======================================================================
// class GraphMFD
// ======================================================================

class DLLEXPORT GraphMFD: public MFD {
public:
	GraphMFD (DWORD w, DWORD h, VESSEL *vessel);
	virtual ~GraphMFD ();
	int AddGraph (void);
	void AddPlot (int g, float *absc, float *data, int ndata, int col, int *ofs = 0);
	void SetRange (int g, int axis, float rmin, float rmax);
	void SetAutoRange (int g, int axis, int p = -1);
	void SetAutoTicks (int g, int axis);
	void SetAxisTitle (int g, int axis, char *title);
	void Plot (HDC hDC, int g, int h0, int h1, const char *title = 0);
	void FindRange (float *d, int ndata, float &dmin, float &dmax) const;

protected:
	int ngraph;  // number of graphs in MFD
	struct GRAPH {
		int nplot; // number of plots in graph
		struct PLOT {
			float *absc;
			float *data;
			int ndata;
			int col;
			int *ofs;
		} *plot;
		float absc_min, absc_max;
		float data_min, data_max;
		float absc_dtick, absc_tickmin, absc_tickscale;
		float data_dtick, data_tickmin, data_tickscale;
		int absc_minortick, data_minortick;
		char absc_title[64], data_title[64];
	} *graph;
};

// ======================================================================
// class CameraMode and subclasses
// ======================================================================

class DLLEXPORT CameraMode {
friend class Camera;
public:
	CameraMode ();
	static CameraMode *Create (char *str);
	virtual void Init (char *str) = 0;
	virtual void Store (char *str) = 0;
	enum Mode { CM_COCKPIT, CM_TRACK, CM_GROUND };
	OBJHANDLE GetTarget() const { return target; }
	virtual Mode GetMode() const = 0;
	virtual void GetDescr (char *str, int len) = 0;
	void SetTarget (OBJHANDLE hTgt);
	void SetFOV (double FOV);
	double GetFOV () const { return fov; }

protected:
	OBJHANDLE target;  // camera target
	double fov;        // field of view [deg]
};

class DLLEXPORT CameraMode_Cockpit: public CameraMode {
public:
	CameraMode_Cockpit ();
	void Init (char *str) {}
	void Store (char *str);
	Mode GetMode() const { return CM_COCKPIT; }
	void GetDescr (char *str, int len);
};

class DLLEXPORT CameraMode_Track: public CameraMode {
public:
	CameraMode_Track ();
	void Init (char *str);
	void Store (char *str);
	Mode GetMode() const { return CM_TRACK; }
	void GetDescr (char *str, int len);
	enum TrackMode { TM_CURRENT, TM_RELATIVE, TM_ABSDIR, TM_GLOBAL, TM_TARGETTOREF, TM_TARGETFROMREF };
	TrackMode GetTrackMode () const { return tmode; }
	void SetTrackMode (TrackMode trackmode, OBJHANDLE refobj = 0);
	OBJHANDLE GetRef() const { return ref; }
	void SetPosition (double rd, double ph, double th);
	void GetPosition (double *rd, double *ph, double *th) const;

protected:
	TrackMode tmode;   // camera track mode
	double reldist;    // distance camera-targets (in units of target size)
	double phi, theta; // camera angle [rad]
	OBJHANDLE ref;     // reference object (for TM_TARGETTOREF and TM_TARGETFROMREF only)
};

class DLLEXPORT CameraMode_Ground: public CameraMode {
public:
	CameraMode_Ground ();
	void Init (char *str);
	void Store (char *str);
	Mode GetMode () const { return CM_GROUND; }
	void GetDescr (char *str, int len);
	void SetPosition (double longitude, double latitude, double altitude, OBJHANDLE hRef = 0);
	void GetPosition (double *longitude, double *latitude, double *altitude, OBJHANDLE *hRef) const;
	void SetOrientation (double ph, double th);
	void GetOrientation (double *ph, double *th) const;
	bool GetTgtLock () const { return tgtlock; }
	OBJHANDLE GetRef() const { return ref; }

protected:
	OBJHANDLE ref;
	double lng, lat;     // camera position on the ground [rad]
	double alt;          // camera height over ground [m]
	double phi, theta;   // camera direction (free mode only)
	bool tgtlock;        // flag for target lock/free mode
};

// ======================================================================
// Orbiter API interface methods
// ======================================================================

OAPIFUNC int        oapiGetOrbiterVersion ();
         int        oapiGetModuleVersion ();
OAPIFUNC HINSTANCE  oapiGetOrbiterInstance ();

OAPIFUNC OBJHANDLE  oapiGetObjectByName (char *name);
OAPIFUNC OBJHANDLE  oapiGetObjectByIndex (int index);
OAPIFUNC DWORD      oapiGetObjectCount ();

OAPIFUNC OBJHANDLE  oapiGetVesselByName (char *name);
OAPIFUNC OBJHANDLE  oapiGetVesselByIndex (int index);
OAPIFUNC DWORD      oapiGetVesselCount ();

OAPIFUNC OBJHANDLE  oapiGetStationByName (char *name);
OAPIFUNC OBJHANDLE  oapiGetStationByIndex (int index);
OAPIFUNC DWORD      oapiGetStationCount ();

OAPIFUNC OBJHANDLE  oapiGetGbodyByName (char *name);
OAPIFUNC OBJHANDLE  oapiGetGbodyByIndex (int index);
OAPIFUNC DWORD      oapiGetGbodyCount ();

OAPIFUNC OBJHANDLE  oapiGetBaseByName (OBJHANDLE hPlanet, char *name);
OAPIFUNC OBJHANDLE  oapiGetBaseByIndex (OBJHANDLE hPlanet, int index);
OAPIFUNC DWORD		oapiGetBaseCount (OBJHANDLE hPlanet);

OAPIFUNC void       oapiGetObjectName (OBJHANDLE hObj, char *name, int n);

OAPIFUNC OBJHANDLE  oapiGetFocusObject ();
OAPIFUNC OBJHANDLE  oapiSetFocusObject (OBJHANDLE hVessel);

OAPIFUNC VESSEL    *oapiGetVesselInterface (OBJHANDLE hVessel);
OAPIFUNC VESSEL    *oapiGetFocusInterface ();

OAPIFUNC OBJHANDLE  oapiCreateVessel (const char *name, const char *classname, const VESSELSTATUS &status);
OAPIFUNC OBJHANDLE  oapiCreateVesselEx (const char *name, const char *classname, const void *status);
OAPIFUNC bool       oapiDeleteVessel (OBJHANDLE hVessel, OBJHANDLE hAlternativeCameraTarget = 0);

OAPIFUNC double     oapiGetSize (OBJHANDLE hObj);
OAPIFUNC double     oapiGetMass (OBJHANDLE hObj);

OAPIFUNC double     oapiGetEmptyMass (OBJHANDLE hVessel);
OAPIFUNC double     oapiGetFuelMass (OBJHANDLE hVessel);
OAPIFUNC void       oapiSetEmptyMass (OBJHANDLE hVessel, double mass);
OAPIFUNC double     oapiGetMaxFuelMass (OBJHANDLE hVessel);
OAPIFUNC double     oapiGetPropellantMaxMass (PROPELLANT_HANDLE ph);
OAPIFUNC double     oapiGetPropellantMass (PROPELLANT_HANDLE ph);
OAPIFUNC PROPELLANT_HANDLE oapiGetPropellantHandle (OBJHANDLE hVessel, DWORD idx);

OAPIFUNC DOCKHANDLE oapiGetDockHandle (OBJHANDLE hVessel, UINT n);
OAPIFUNC OBJHANDLE  oapiGetDockStatus (DOCKHANDLE dock);

OAPIFUNC void       oapiGetGlobalPos (OBJHANDLE hObj, VECTOR3 *pos);
OAPIFUNC void       oapiGetGlobalVel (OBJHANDLE hObj, VECTOR3 *vel);
OAPIFUNC void       oapiGetFocusGlobalPos (VECTOR3 *pos);
OAPIFUNC void       oapiGetFocusGlobalVel (VECTOR3 *vel);

OAPIFUNC void       oapiGetRelativePos (OBJHANDLE hObj, OBJHANDLE hRef, VECTOR3 *pos);
OAPIFUNC void       oapiGetRelativeVel (OBJHANDLE hObj, OBJHANDLE hRef, VECTOR3 *vel);
OAPIFUNC void       oapiGetFocusRelativePos (OBJHANDLE hRef, VECTOR3 *pos);
OAPIFUNC void       oapiGetFocusRelativeVel (OBJHANDLE hRef, VECTOR3 *vel);

OAPIFUNC void       oapiGetBarycentre (OBJHANDLE hObj, VECTOR3 *bary);

OAPIFUNC BOOL       oapiGetAltitude (OBJHANDLE hVessel, double *alt);
OAPIFUNC BOOL       oapiGetPitch (OBJHANDLE hVessel, double *pitch);
OAPIFUNC BOOL       oapiGetBank (OBJHANDLE hVessel, double *bank);
OAPIFUNC BOOL       oapiGetHeading (OBJHANDLE hVessel, double *heading);
OAPIFUNC BOOL       oapiGetEquPos (OBJHANDLE hVessel, double *longitude, double *latitude, double *radius);
OAPIFUNC BOOL       oapiGetFocusAltitude (double *alt);
OAPIFUNC BOOL       oapiGetFocusPitch (double *pitch);
OAPIFUNC BOOL       oapiGetFocusBank (double *bank);
OAPIFUNC BOOL       oapiGetFocusHeading (double *heading);
OAPIFUNC BOOL       oapiGetFocusEquPos (double *longitude, double *latitude, double *radius);

OAPIFUNC BOOL       oapiGetAirspeed (OBJHANDLE hVessel, double *airspeed);
OAPIFUNC BOOL       oapiGetAirspeedVector (OBJHANDLE hVessel, VECTOR3 *speedvec);
OAPIFUNC BOOL       oapiGetShipAirspeedVector (OBJHANDLE hVessel, VECTOR3 *speedvec);
OAPIFUNC BOOL       oapiGetFocusAirspeed (double *airspeed);
OAPIFUNC BOOL       oapiGetFocusAirspeedVector (VECTOR3 *speedvec);
OAPIFUNC BOOL       oapiGetFocusShipAirspeedVector (VECTOR3 *speedvec);

OAPIFUNC void       oapiGetAtmPressureDensity (OBJHANDLE hVessel, double *pressure, double *density);
OAPIFUNC void       oapiGetFocusAtmPressureDensity (double *pressure, double *density);

OAPIFUNC void       oapiGetEngineStatus (OBJHANDLE hVessel, ENGINESTATUS *es);
OAPIFUNC void       oapiGetFocusEngineStatus (ENGINESTATUS *es);
OAPIFUNC void       oapiSetEngineLevel (OBJHANDLE hVessel, ENGINETYPE engine, double level);
OAPIFUNC SURFHANDLE oapiRegisterExhaustTexture (char *name);
OAPIFUNC SURFHANDLE oapiRegisterReentryTexture (char *name);
OAPIFUNC SURFHANDLE oapiRegisterParticleTexture (char *name);

OAPIFUNC int        oapiGetAttitudeMode (OBJHANDLE hVessel);
OAPIFUNC int        oapiToggleAttitudeMode (OBJHANDLE hVessel);
OAPIFUNC bool       oapiSetAttitudeMode (OBJHANDLE hVessel, int mode);
OAPIFUNC int        oapiGetFocusAttitudeMode ();
OAPIFUNC int        oapiToggleFocusAttitudeMode ();
OAPIFUNC bool       oapiSetFocusAttitudeMode (int mode);

OAPIFUNC void       oapiSetShowGrapplePoints (bool show);
OAPIFUNC bool       oapiGetShowGrapplePoints ();

OAPIFUNC double     oapiGetSimTime ();
OAPIFUNC double     oapiGetSimStep ();
OAPIFUNC double     oapiGetSysTime ();
OAPIFUNC double     oapiGetSysStep ();
OAPIFUNC double     oapiGetSimMJD ();
OAPIFUNC double     oapiGetFrameRate ();
OAPIFUNC double     oapiTime2MJD (double t);
OAPIFUNC double     oapiGetTimeAcceleration ();
OAPIFUNC void       oapiSetTimeAcceleration (double warp);
OAPIFUNC bool       oapiAcceptDelayedKey (char key, double interval);

// Aerodynamics helper functions
OAPIFUNC double     oapiGetInducedDrag (double cl, double A, double e);
OAPIFUNC double     oapiGetWaveDrag (double M, double M1, double M2, double M3, double cmax);

// Camera functions
OAPIFUNC bool		oapiCameraInternal ();
OAPIFUNC int        oapiCameraMode ();
OAPIFUNC int        oapiCockpitMode ();
OAPIFUNC OBJHANDLE  oapiCameraTarget ();
OAPIFUNC void		oapiCameraGlobalPos (VECTOR3 *gpos);
OAPIFUNC void		oapiCameraGlobalDir (VECTOR3 *gdir);
OAPIFUNC double		oapiCameraTargetDist ();
OAPIFUNC double		oapiCameraAzimuth ();
OAPIFUNC double		oapiCameraPolar ();
OAPIFUNC double		oapiCameraAperture ();
OAPIFUNC void		oapiCameraSetAperture (double aperture);
OAPIFUNC void		oapiCameraScaleDist (double dscale);
OAPIFUNC void		oapiCameraRotAzimuth (double dazimuth);
OAPIFUNC void		oapiCameraRotPolar (double dpolar);
OAPIFUNC void       oapiCameraSetCockpitDir (double polar, double azimuth, bool transition = false);
OAPIFUNC void		oapiCameraAttach (OBJHANDLE hObj, int mode);

// Functions for planetary bodies
OAPIFUNC double     oapiGetPlanetPeriod (OBJHANDLE hPlanet);
OAPIFUNC double		oapiGetPlanetObliquity (OBJHANDLE hPlanet);
OAPIFUNC double		oapiGetPlanetTheta (OBJHANDLE hPlanet);
OAPIFUNC void		oapiGetPlanetObliquityMatrix (OBJHANDLE hPlanet, MATRIX3 *mat);
OAPIFUNC double		oapiGetPlanetCurrentRotation (OBJHANDLE hPlanet);
OAPIFUNC bool		oapiPlanetHasAtmosphere (OBJHANDLE hPlanet);
OAPIFUNC void       oapiGetPlanetAtmParams (OBJHANDLE hPlanet, double rad, ATMPARAM *prm);
OAPIFUNC const ATMCONST *oapiGetPlanetAtmConstants (OBJHANDLE hPlanet);
OAPIFUNC DWORD      oapiGetPlanetJCoeffCount (OBJHANDLE hPlanet);
OAPIFUNC double     oapiGetPlanetJCoeff (OBJHANDLE hPlanet, DWORD n);

// Surface base interface
OAPIFUNC void       oapiGetBaseEquPos (OBJHANDLE hBase, double *lng, double *lat, double *rad = 0);
OAPIFUNC DWORD		oapiGetBasePadCount (OBJHANDLE hBase);
OAPIFUNC bool       oapiGetBasePadEquPos (OBJHANDLE hBase, DWORD pad, double *lng, double *lat, double *rad = 0);
OAPIFUNC bool       oapiGetBasePadStatus (OBJHANDLE hBase, DWORD pad, int *status);
OAPIFUNC NAVHANDLE  oapiGetBasePadNav (OBJHANDLE hBase, DWORD pad);

// Navigation radio transmitter functions
OAPIFUNC void       oapiGetNavPos (NAVHANDLE hNav, VECTOR3 *gpos);
OAPIFUNC DWORD		oapiGetNavChannel (NAVHANDLE hNav);
OAPIFUNC float		oapiGetNavFreq (NAVHANDLE hNav);
OAPIFUNC float      oapiGetNavRange (NAVHANDLE hNav);
OAPIFUNC bool		oapiNavInRange (NAVHANDLE hNav, const VECTOR3 &gpos);

// Mesh functions
OAPIFUNC MESHHANDLE oapiLoadMesh (const char *fname);
OAPIFUNC const MESHHANDLE oapiLoadMeshGlobal (const char *fname);
OAPIFUNC void       oapiDeleteMesh (MESHHANDLE hMesh);
OAPIFUNC SURFHANDLE oapiGetTextureHandle (MESHHANDLE hMesh, DWORD texidx);
OAPIFUNC bool       oapiSetTexture (MESHHANDLE hMesh, DWORD texidx, SURFHANDLE tex);
OAPIFUNC DWORD      oapiMeshGroupCount (MESHHANDLE hMesh);
OAPIFUNC MESHGROUP *oapiMeshGroup (MESHHANDLE hMesh, DWORD idx);

// Particle functions
OAPIFUNC void       oapiParticleSetLevelRef (PSTREAM_HANDLE ph, double *lvl);

// HUD, MFD and panel functions
OAPIFUNC bool       oapiSetHUDMode (int mode);
OAPIFUNC int        oapiGetHUDMode ();
OAPIFUNC void       oapiToggleHUDColour ();
OAPIFUNC void       oapiIncHUDIntensity ();
OAPIFUNC void       oapiDecHUDIntensity ();
OAPIFUNC void       oapiOpenMFD (int mode, int mfd);
OAPIFUNC void       oapiToggleMFD_on (int mfd);
OAPIFUNC int	    oapiGetMFDMode (int mfd);
OAPIFUNC int        oapiBroadcastMFDMessage (int mode, int msg, void *data);
OAPIFUNC int        oapiSendMFDKey (int mfd, DWORD key);
OAPIFUNC bool       oapiProcessMFDButton (int mfd, int bt, int event);
OAPIFUNC const char *oapiMFDButtonLabel (int mfd, int bt);
OAPIFUNC void       oapiRegisterMFD (int mfd, const MFDSPEC &spec);
OAPIFUNC void       oapiRegisterPanelBackground (HBITMAP hBmp, DWORD flag = PANEL_ATTACH_BOTTOM|PANEL_MOVEOUT_BOTTOM,
				        DWORD ck = (DWORD)-1);
OAPIFUNC void       oapiRegisterPanelArea (int id, const RECT &pos, int draw_event = PANEL_REDRAW_NEVER,
					    int mouse_event = PANEL_MOUSE_IGNORE, int bkmode = PANEL_MAP_NONE);
OAPIFUNC void       oapiSetPanelNeighbours (int left, int right, int top, int bottom);
OAPIFUNC void       oapiTriggerPanelRedrawArea (int panel_id, int area_id);
OAPIFUNC void       oapiTriggerRedrawArea (int panel_id, int vc_id, int area_id);
OAPIFUNC bool       oapiBltPanelAreaBackground (int area_id, SURFHANDLE surf);
OAPIFUNC int        oapiSwitchPanel (int direction);
OAPIFUNC int        oapiSetPanel (int panel_id);
OAPIFUNC HDC        oapiGetDC (SURFHANDLE surf);
OAPIFUNC void       oapiReleaseDC (SURFHANDLE surf, HDC hDC);
OAPIFUNC DWORD      oapiGetColour (DWORD red, DWORD green, DWORD blue);
OAPIFUNC SURFHANDLE oapiCreateSurface (int width, int height);
OAPIFUNC SURFHANDLE oapiCreateSurface (HBITMAP hBmp, bool release_bmp = true);
OAPIFUNC SURFHANDLE oapiCreateTextureSurface (int width, int height);
OAPIFUNC void       oapiDestroySurface (SURFHANDLE surf);
OAPIFUNC void       oapiClearSurface (SURFHANDLE surf, DWORD col = 0);
OAPIFUNC void       oapiSetSurfaceColourKey (SURFHANDLE surf, DWORD ck);
OAPIFUNC void       oapiClearSurfaceColourKey (SURFHANDLE surf);
OAPIFUNC void       oapiBlt (SURFHANDLE tgt, SURFHANDLE src, int tgtx, int tgty, int srcx, int srcy, int w, int h, DWORD ck = SURF_NO_CK);
OAPIFUNC void       oapiBlt (SURFHANDLE tgt, SURFHANDLE src, RECT *tgtr, RECT *srcr, DWORD ck = SURF_NO_CK, DWORD rotate = SURF_NO_ROTATION);
OAPIFUNC void       oapiColourFill (SURFHANDLE tgt, DWORD fillcolor, int tgtx = 0, int tgty = 0, int w = 0, int h = 0);

// Custom MFD mode definition
OAPIFUNC int        oapiRegisterMFDMode (MFDMODESPEC &spec);
OAPIFUNC bool       oapiUnregisterMFDMode (int mode);
OAPIFUNC void       oapiDisableMFDMode (int mode);
OAPIFUNC int        oapiGetMFDModeSpec (char *name, MFDMODESPEC **spec = 0);

// virtual cockpit functions
OAPIFUNC void       oapiVCRegisterMFD (int mfd, const VCMFDSPEC *spec);
OAPIFUNC void       oapiVCRegisterArea (int id, const RECT &tgtrect, int draw_event, int mouse_event, int bkmode, SURFHANDLE tgt);
OAPIFUNC void       oapiVCRegisterArea (int id, int draw_event, int mouse_event);
OAPIFUNC void       oapiVCSetAreaClickmode_Spherical (int id, const VECTOR3 &cnt, double rad);
OAPIFUNC void       oapiVCSetAreaClickmode_Quadrilateral (int id, const VECTOR3 &p1, const VECTOR3 &p2, const VECTOR3 &p3, const VECTOR3 &p4);
OAPIFUNC void       oapiVCTriggerRedrawArea (int vc_id, int area_id);
OAPIFUNC void       oapiVCRegisterHUD (const VCHUDSPEC *spec);

// Customisation - custom menu, dialogs
typedef void (*CustomFunc)(void *context);
OAPIFUNC DWORD      oapiRegisterCustomCmd (char *label, char *desc, CustomFunc func, void *context);
OAPIFUNC bool       oapiUnregisterCustomCmd (int cmdId);

OAPIFUNC HWND       oapiOpenDialog (HINSTANCE hDLLInst, int resourceId, DLGPROC msgProc, void *context = 0);
OAPIFUNC HWND       oapiFindDialog (HINSTANCE hDLLInst, int resourceId);
OAPIFUNC void       oapiCloseDialog (HWND hDlg);
OAPIFUNC void      *oapiGetDialogContext (HWND hDlg);
OAPIFUNC BOOL       oapiDefDialogProc (HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam);

// File IO
OAPIFUNC void       oapiWriteLine (FILEHANDLE file, char *line);
OAPIFUNC void       oapiWriteScenario_string  (FILEHANDLE scn, char *item, char *string);
OAPIFUNC void       oapiWriteScenario_int     (FILEHANDLE scn, char *item, int i);
OAPIFUNC void       oapiWriteScenario_float   (FILEHANDLE scn, char *item, double d);
OAPIFUNC void       oapiWriteScenario_vec     (FILEHANDLE scn, char *item, const VECTOR3 &vec);
OAPIFUNC bool	    oapiReadScenario_nextline (FILEHANDLE scn, char *&line);
OAPIFUNC bool       oapiReadItem_string       (FILEHANDLE f, char *item, char *string);
OAPIFUNC bool       oapiReadItem_float        (FILEHANDLE f, char *item, double &d);
OAPIFUNC bool       oapiReadItem_int          (FILEHANDLE f, char *item, int &i);
OAPIFUNC bool       oapiReadItem_bool         (FILEHANDLE f, char *item, bool &b);
OAPIFUNC bool       oapiReadItem_vec          (FILEHANDLE f, char *item, VECTOR3 &vec);

// User Input
OAPIFUNC void       oapiOpenInputBox (char *title, bool (*Clbk)(void*,char*,void*), char *buf = 0, int vislen = 20, void *usrdata = 0);

OAPIFUNC char      *oapiDebugString ();

// ======================================================================
// keyboard key identifiers
// ======================================================================

#define OAPI_KEY_ESCAPE			0x01
#define OAPI_KEY_1				0x02
#define OAPI_KEY_2				0x03
#define OAPI_KEY_3				0x04
#define OAPI_KEY_4				0x05
#define OAPI_KEY_5				0x06
#define OAPI_KEY_6				0x07
#define OAPI_KEY_7				0x08
#define OAPI_KEY_8				0x09
#define OAPI_KEY_9				0x0A
#define OAPI_KEY_0				0x0B
#define OAPI_KEY_MINUS			0x0C  // on main keyboard
#define OAPI_KEY_EQUALS			0x0D
#define OAPI_KEY_BACK			0x0E  // backspace
#define OAPI_KEY_TAB			0x0F
#define OAPI_KEY_Q				0x10
#define OAPI_KEY_W				0x11
#define OAPI_KEY_E				0x12
#define OAPI_KEY_R				0x13
#define OAPI_KEY_T				0x14
#define OAPI_KEY_Y				0x15
#define OAPI_KEY_U				0x16
#define OAPI_KEY_I				0x17
#define OAPI_KEY_O				0x18
#define OAPI_KEY_P				0x19
#define OAPI_KEY_LBRACKET		0x1A
#define OAPI_KEY_RBRACKET		0x1B
#define OAPI_KEY_RETURN			0x1C  // Enter on main keyboard
#define OAPI_KEY_LCONTROL		0x1D
#define OAPI_KEY_A				0x1E
#define OAPI_KEY_S				0x1F
#define OAPI_KEY_D				0x20
#define OAPI_KEY_F				0x21
#define OAPI_KEY_G				0x22
#define OAPI_KEY_H				0x23
#define OAPI_KEY_J				0x24
#define OAPI_KEY_K				0x25
#define OAPI_KEY_L				0x26
#define OAPI_KEY_SEMICOLON		0x27
#define OAPI_KEY_APOSTROPHE		0x28
#define OAPI_KEY_GRAVE			0x29  // accent grave
#define OAPI_KEY_LSHIFT			0x2A
#define OAPI_KEY_BACKSLASH		0x2B
#define OAPI_KEY_Z				0x2C
#define OAPI_KEY_X				0x2D
#define OAPI_KEY_C				0x2E
#define OAPI_KEY_V				0x2F
#define OAPI_KEY_B				0x30
#define OAPI_KEY_N				0x31
#define OAPI_KEY_M				0x32
#define OAPI_KEY_COMMA			0x33
#define OAPI_KEY_PERIOD			0x34  // . on main keyboard
#define OAPI_KEY_SLASH			0x35  // / on main keyboard
#define OAPI_KEY_RSHIFT			0x36
#define OAPI_KEY_MULTIPLY		0x37  // * on numeric keypad
#define OAPI_KEY_LALT			0x38  // left Alt
#define OAPI_KEY_SPACE			0x39
#define OAPI_KEY_CAPITAL		0x3A  // caps lock key
#define OAPI_KEY_F1				0x3B
#define OAPI_KEY_F2				0x3C
#define OAPI_KEY_F3				0x3D
#define OAPI_KEY_F4				0x3E
#define OAPI_KEY_F5				0x3F
#define OAPI_KEY_F6				0x40
#define OAPI_KEY_F7				0x41
#define OAPI_KEY_F8				0x42
#define OAPI_KEY_F9				0x43
#define OAPI_KEY_F10			0x44
#define OAPI_KEY_NUMLOCK		0x45
#define OAPI_KEY_SCROLL			0x46  // Scroll lock
#define OAPI_KEY_NUMPAD7		0x47
#define OAPI_KEY_NUMPAD8		0x48
#define OAPI_KEY_NUMPAD9		0x49
#define OAPI_KEY_SUBTRACT		0x4A  // - on numeric keypad
#define OAPI_KEY_NUMPAD4		0x4B
#define OAPI_KEY_NUMPAD5		0x4C
#define OAPI_KEY_NUMPAD6		0x4D
#define OAPI_KEY_ADD			0x4E  // + on numeric keypad
#define OAPI_KEY_NUMPAD1		0x4F
#define OAPI_KEY_NUMPAD2		0x50
#define OAPI_KEY_NUMPAD3		0x51
#define OAPI_KEY_NUMPAD0		0x52
#define OAPI_KEY_DECIMAL		0x53  // . on numeric keypad
#define OAPI_KEY_OEM_102		0x56  // | < > on UK/German keyboards
#define OAPI_KEY_F11			0x57
#define OAPI_KEY_F12			0x58
#define OAPI_KEY_NUMPADENTER	0x9C  // Enter on numeric keypad
#define OAPI_KEY_RCONTROL		0x9D  // right Control key
#define OAPI_KEY_DIVIDE			0xB5  // / on numeric keypad
#define OAPI_KEY_RALT           0xB8  // right Alt
#define OAPI_KEY_HOME           0xC7  // Home on cursor keypad
#define OAPI_KEY_UP             0xC8  // up-arrow on cursor keypad
#define OAPI_KEY_PRIOR          0xC9  // PgUp on cursor keypad
#define OAPI_KEY_LEFT           0xCB  // left-arrow on cursor keypad
#define OAPI_KEY_RIGHT          0xCD  // right-arrow on cursor keypad
#define OAPI_KEY_END            0xCF  // End on cursor keypad
#define OAPI_KEY_DOWN           0xD0  // down-arrow on cursor keypad
#define OAPI_KEY_NEXT           0xD1  // PgDn on cursor keypad
#define OAPI_KEY_INSERT         0xD2  // Insert on cursor keypad
#define OAPI_KEY_DELETE         0xD3  // Delete on cursor keypad

#define KEYDOWN(buf,key) (buf[key] & 0x80)
#define RESETKEY(buf,key) (buf[key] = 0)

#define KEYMOD_LSHIFT(buf)   (KEYDOWN(buf,OAPI_KEY_LSHIFT))
#define KEYMOD_RSHIFT(buf)   (KEYDOWN(buf,OAPI_KEY_RSHIFT))
#define KEYMOD_SHIFT(buf)    (KEYMOD_LSHIFT(buf) || KEYMOD_RSHIFT(buf))
#define KEYMOD_LCONTROL(buf) (KEYDOWN(buf,OAPI_KEY_LCONTROL))
#define KEYMOD_RCONTROL(buf) (KEYDOWN(buf,OAPI_KEY_RCONTROL))
#define KEYMOD_CONTROL(buf)  (KEYMOD_LCONTROL(buf) || KEYMOD_RCONTROL(buf))
#define KEYMOD_LALT(buf)     (KEYDOWN(buf,OAPI_KEY_LALT))
#define KEYMOD_RALT(buf)     (KEYDOWN(buf,OAPI_KEY_RALT))
#define KEYMOD_ALT(buf)      (KEYMOD_LALT(buf) || KEYMOD_RALT(buf))

// ======================================================================
// Logical key ids
// ======================================================================

#define OAPI_LKEY_CockpitRotateLeft  0
#define OAPI_LKEY_CockpitRotateRight 1
#define OAPI_LKEY_CockpitRotateUp    2
#define OAPI_LKEY_CockpitRotateDown  3
#define OAPI_LKEY_CockpitDontLean    4
#define OAPI_LKEY_CockpitLeanForward 5
#define OAPI_LKEY_CockpitLeanLeft    6
#define OAPI_LKEY_CockpitLeanRight   7
#define OAPI_LKEY_CockpitResetCam    8
#define OAPI_LKEY_PanelShiftLeft     9
#define OAPI_LKEY_PanelShiftRight   10
#define OAPI_LKEY_PanelShiftUp      11
#define OAPI_LKEY_PanelShiftDown    12
#define OAPI_LKEY_PanelSwitchLeft   13
#define OAPI_LKEY_PanelSwitchRight  14
#define OAPI_LKEY_PanelSwitchUp     15
#define OAPI_LKEY_PanelSwitchDown   16
#define OAPI_LKEY_TrackRotateLeft   17
#define OAPI_LKEY_TrackRotateRight  18
#define OAPI_LKEY_TrackRotateUp     19
#define OAPI_LKEY_TrackRotateDown   20
#define OAPI_LKEY_TrackAdvance      21
#define OAPI_LKEY_TrackRetreat      22
#define OAPI_LKEY_GroundTiltLeft    23
#define OAPI_LKEY_GroundTiltRight   24
#define OAPI_LKEY_GroundTiltUp      25
#define OAPI_LKEY_GroundTiltDown    26
#define OAPI_LKEY_IncMainThrust     27  //  +-
#define OAPI_LKEY_DecMainThrust     28  //  |
#define OAPI_LKEY_KillMainRetro     29  //  |
#define OAPI_LKEY_FullMainThrust    30  //  |   Main/retro/hover engine control
#define OAPI_LKEY_FullRetroThrust   31  //  |
#define OAPI_LKEY_IncHoverThrust    32  //  |
#define OAPI_LKEY_DecHoverThrust    33  //  +-
#define OAPI_LKEY_RCSPitchUp        34  //  +-
#define OAPI_LKEY_RCSPitchDown      35  //  |
#define OAPI_LKEY_RCSYawLeft        36  //  |   Reaction control
#define OAPI_LKEY_RCSYawRight       37  //  |   (rotational mode)
#define OAPI_LKEY_RCSBankLeft       38  //  |
#define OAPI_LKEY_RCSBankRight      39  //  +-
#define OAPI_LKEY_RCSUp             40  //  +-
#define OAPI_LKEY_RCSDown           41  //  |
#define OAPI_LKEY_RCSLeft           42  //  |   Reaction control
#define OAPI_LKEY_RCSRight          43  //  |   (linear mode)
#define OAPI_LKEY_RCSForward        44  //  |
#define OAPI_LKEY_RCSBack           45  //  +-
#define OAPI_LKEY_LPRCSPitchUp      46  //  +-
#define OAPI_LKEY_LPRCSPitchDown    47  //  |
#define OAPI_LKEY_LPRCSYawLeft      48  //  |   Reaction control - low power
#define OAPI_LKEY_LPRCSYawRight     49  //  |   (rotational mode)
#define OAPI_LKEY_LPRCSBankLeft     50  //  |
#define OAPI_LKEY_LPRCSBankRight    51  //  +-
#define OAPI_LKEY_LPRCSUp           52  //  +-
#define OAPI_LKEY_LPRCSDown         53  //  |
#define OAPI_LKEY_LPRCSLeft         54  //  |   Reaction control - low power
#define OAPI_LKEY_LPRCSRight        55  //  |   (linear mode)
#define OAPI_LKEY_LPRCSForward      56  //  |
#define OAPI_LKEY_LPRCSBack         57  //  +-
#define OAPI_LKEY_NMHoldAltitude    58  //  +-
#define OAPI_LKEY_NMHLevel          59  //  |
#define OAPI_LKEY_NMPrograde        60  //  |
#define OAPI_LKEY_NMRetrograde      61  //  |   Navigation computer modes
#define OAPI_LKEY_NMNormal          62  //  |
#define OAPI_LKEY_NMAntinormal      63  //  |
#define OAPI_LKEY_NMKillrot         64  //  +-
#define OAPI_LKEY_IncElevatorTrim   65
#define OAPI_LKEY_DecElevatorTrim   66
#define OAPI_LKEY_WheelbrakeLeft    67
#define OAPI_LKEY_WheelbrakeRight   68
#define OAPI_LKEY_HUD               69
#define OAPI_LKEY_HUDMode           70
#define OAPI_LKEY_HUDReference      71
#define OAPI_LKEY_HUDTarget         72
#define OAPI_LKEY_HUDColour         73
#define OAPI_LKEY_IncSimSpeed       74
#define OAPI_LKEY_DecSimSpeed       75
#define OAPI_LKEY_IncFOV            76
#define OAPI_LKEY_DecFOV            77
#define OAPI_LKEY_MainMenu          78
#define OAPI_LKEY_DlgHelp           79
#define OAPI_LKEY_DlgCamera         80
#define OAPI_LKEY_DlgSimspeed       81
#define OAPI_LKEY_DlgCustomCmd      82
#define OAPI_LKEY_DlgPlanetarium    83
#define OAPI_LKEY_DlgInfo           84
#define OAPI_LKEY_DlgMap            85
#define OAPI_LKEY_DlgNavaid         86
#define OAPI_LKEY_ToggleInfo        87
#define OAPI_LKEY_ToggleFPS         88
#define OAPI_LKEY_ToggleCamInternal 89
#define OAPI_LKEY_ToggleTrackMode   90
#define OAPI_LKEY_TogglePanelMode   91
#define OAPI_LKEY_TogglePlanetarium 92
#define LKEY_COUNT 93

// ======================================================================
// Some helper functions
// ======================================================================

inline VECTOR3 _V(double x, double y, double z)
{
	VECTOR3 vec = {x,y,z}; return vec;
}

inline void veccpy (VECTOR3 &a, const VECTOR3 &b)
{
	a.x = b.x;
	a.y = b.y;
	a.z = b.z;
}

inline VECTOR3 operator+ (const VECTOR3 &a, const VECTOR3 &b)
{
	VECTOR3 c;
	c.x = a.x+b.x;
	c.y = a.y+b.y;
	c.z = a.z+b.z;
	return c;
}

inline VECTOR3 operator- (const VECTOR3 &a, const VECTOR3 &b)
{
	VECTOR3 c;
	c.x = a.x-b.x;
	c.y = a.y-b.y;
	c.z = a.z-b.z;
	return c;
}

inline VECTOR3 operator* (const VECTOR3 &a, const double f)
{
	VECTOR3 c;
	c.x = a.x*f;
	c.y = a.y*f;
	c.z = a.z*f;
	return c;
}

inline VECTOR3 &operator+= (VECTOR3 &a, const VECTOR3 &b)
{
	a.x += b.x;
	a.y += b.y;
	a.z += b.z;
	return a;
}

inline VECTOR3 &operator-= (VECTOR3 &a, const VECTOR3 &b)
{
	a.x -= b.x;
	a.y -= b.y;
	a.z -= b.z;
	return a;
}

inline VECTOR3 &operator*= (VECTOR3 &a, const double f)
{
	a.x *= f;
	a.y *= f;
	a.z *= f;
	return a;
}

inline VECTOR3 &operator/= (VECTOR3 &a, const double f)
{
	a.x /= f;
	a.y /= f;
	a.z /= f;
	return a;
}

inline double dotp (const VECTOR3& a, const VECTOR3 &b)
{
	return a.x*b.x + a.y*b.y + a.z*b.z;
}

inline VECTOR3 crossp (const VECTOR3 &a, const VECTOR3 &b)
{
	return _V(a.y*b.z - b.y*a.z, a.z*b.x - b.z*a.x, a.x*b.y - b.x*a.y);
}

inline double length (const VECTOR3 &a)
{
	return sqrt (a.x*a.x + a.y*a.y + a.z*a.z);
}

inline double dist (const VECTOR3 &a, const VECTOR3 &b)
{
	return length (a-b);
}

inline RECT _R (int left, int top, int right, int bottom)
{
	RECT r = { left, top, right, bottom }; return r;
}

inline VECTOR3 POINTERTOREF (VECTOR3 *p)
{
	VECTOR3 v;
	v.x = DBL_MAX;            // flag
	*((VECTOR3**)&v.z) = p;   // address
	v.z = 0.0;
	return v;
}

// ======================================================================
// Internal data structures
// ======================================================================

#ifdef ORBITER_MODULE
void dummy();
void calldummy () { dummy(); }
#endif

#endif // !__ORBITERAPI_H